#!tools/lua52.exe
-- http://files.luaforge.net/releases/luabinaries/aLua5.2-work2/Executables

-- mods:
-- http://computer.freepage.de/nofuture/

local thomson = {div=4, compress=true}

function ModFile(mod)
	-- converted from https://github.com/gasman/jsmodplayer/blob/master/modfile.js
	local function substr(str,pos,len) 
		return str:sub(pos+1,pos+len)
	end
	local function trimNulls(str)
		return str:gsub("%z*$","")
	end
	local function cleanString(str)
		return trimNulls(str):gsub("%c",""):gsub("[^%w%p%s]","?")
	end
	
	local function getWord(str, pos)
		local a,b = str:byte(pos+1,pos+2)
		return a*256+b
	end
	local function getByte(str, pos)
		return str:byte(pos+1)
	end

	local this = {}
	this.data = mod
	this.samples = {}
	this.positions = {}
	this.patternCount = 0
	this.patterns = {}

	this.title = cleanString(substr(mod,0,20))
	
	--[[
	TODO: distinguish 15-sample and 31-sample mods. Currently assuming 31-sample.
	"Check the bytes at location 471 in the file. If there is text there (ASCII
	$20-$7E (32-126)), then you can probably assume it's a 31-instrument file.
	Otherwise, it's an older 15 instrument file." */
	--]]
	this.sampleCount = getByte(mod,471)>=32 and getByte(mod,471)<=126 and 31 or 15

	for i=1,this.sampleCount do
		local sampleInfo = substr(mod, 20+(i-1)*30, 30)
		local sampleName = cleanString(substr(sampleInfo, 0, 22))
		this.samples[i] = {
		    no =  i-1,
			name = sampleName,
			length = getWord(sampleInfo, 22) * 2,
			finetune = getByte(sampleInfo, 24),
			volume = getByte(sampleInfo, 25),
			repeatOffset = getWord(sampleInfo, 26) * 2,
			repeatLength = getWord(sampleInfo, 28) * 2
		}
	end
	
	this.positionCount = getByte(mod,950)
	this.positionLoopPoint = getByte(mod, 951)+1

	for i=1,this.positionCount do
		this.positions[i] = getByte(mod, 951+i)+1
		this.patternCount = math.max(this.patternCount, this.positions[i])
	end

	local identifier = substr(mod,1080,4)
	local channelCountByIdentifier = {
	     ['TDZ1'] = 1, ['1CHN'] = 1, ['TDZ2'] = 2, ['2CHN'] = 2, ['TDZ3'] = 3, ['3CHN'] = 3,
	     ['M.K.'] = 4, ['FLT4'] = 4, ['M!K!'] = 4, ['4CHN'] = 4, ['TDZ4'] = 4, ['5CHN'] = 5, ['TDZ5'] = 5,
	     ['6CHN'] = 6, ['TDZ6'] = 6, ['7CHN'] = 7, ['TDZ7'] = 7, ['8CHN'] = 8, ['TDZ8'] = 8, ['OCTA'] = 8, ['CD81'] = 8,
	     ['9CHN'] = 9, ['TDZ9'] = 9,
	     ['10CH'] = 10, ['11CH'] = 11, ['12CH'] = 12, ['13CH'] = 13, ['14CH'] = 14, ['15CH'] = 15, ['16CH'] = 16, ['17CH'] = 17,
	     ['18CH'] = 18, ['19CH'] = 19, ['20CH'] = 20, ['21CH'] = 21, ['22CH'] = 22, ['23CH'] = 23, ['24CH'] = 24, ['25CH'] = 25,
	     ['26CH'] = 26, ['27CH'] = 27, ['28CH'] = 28, ['29CH'] = 29, ['30CH'] = 30, ['31CH'] = 31, ['32CH'] = 32
	}
	this.channelCount = channelCountByIdentifier[identifier] or 4
	this.identifier = identifier

	local patternOffset = 1084
	for pat=1,this.patternCount do
		this.patterns[pat] = {no=pat};
		for row=1,64 do
			this.patterns[pat][row] = {};
			for chan=1,this.channelCount do
				local b0 = getByte(mod,patternOffset + 0)
				local b1 = getByte(mod,patternOffset + 1)
				local b2 = getByte(mod,patternOffset + 2) 
				local b3 = getByte(mod,patternOffset + 3)
				this.patterns[pat][row][chan] = {
					sample = math.floor(b0/16)*16 + math.floor(b2/16),
					period = (b0 % 16)*256 + b1,
					effect = b2 % 16,
					effectParameter = b3
				}
				patternOffset = patternOffset + 4
			end
		end
	end

	local sampleOffset = patternOffset
	for s=1,this.sampleCount do	
		this.samples[s].startOffset = sampleOffset;
		this.samples[s].getSample = function (pos) 
			if pos>=this.samples[s].length then pos=this.samples[s].length-1 end
			local b = getByte(mod, this.samples[s].startOffset+pos) or 0
			-- return b -- unsigned
			return ((b+128)%256)-128 -- signed
		end
		sampleOffset = sampleOffset + this.samples[s].length
	end
	
	return this
end

--[[
ModPeriodTable[ft][n] = the period to use for note number n at finetune value ft.
Finetune values are in twos-complement, i.e. [0,1,2,3,4,5,6,7,-8,-7,-6,-5,-4,-3,-2,-1]
The first table is used to generate a reverse lookup table, to find out the note number
for a period given in the MOD file.
--]]
local ModPeriodTable = {
	{1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016, 960 , 906,
	 856 , 808 , 762 , 720 , 678 , 640 , 604 , 570 , 538 , 508 , 480 , 453,
	 428 , 404 , 381 , 360 , 339 , 320 , 302 , 285 , 269 , 254 , 240 , 226,
	 214 , 202 , 190 , 180 , 170 , 160 , 151 , 143 , 135 , 127 , 120 , 113,
	 107 , 101 , 95  , 90  , 85  , 80  , 75  , 71  , 67  , 63  , 60  , 56 },
	{1700, 1604, 1514, 1430, 1348, 1274, 1202, 1134, 1070, 1010, 954 , 900,
	 850 , 802 , 757 , 715 , 674 , 637 , 601 , 567 , 535 , 505 , 477 , 450,
	 425 , 401 , 379 , 357 , 337 , 318 , 300 , 284 , 268 , 253 , 239 , 225,
	 213 , 201 , 189 , 179 , 169 , 159 , 150 , 142 , 134 , 126 , 119 , 113,
	 106 , 100 , 94  , 89  , 84  , 79  , 75  , 71  , 67  , 63  , 59  , 56 },
	{1688, 1592, 1504, 1418, 1340, 1264, 1194, 1126, 1064, 1004, 948 , 894,
	 844 , 796 , 752 , 709 , 670 , 632 , 597 , 563 , 532 , 502 , 474 , 447,
	 422 , 398 , 376 , 355 , 335 , 316 , 298 , 282 , 266 , 251 , 237 , 224,
	 211 , 199 , 188 , 177 , 167 , 158 , 149 , 141 , 133 , 125 , 118 , 112,
	 105 , 99  , 94  , 88  , 83  , 79  , 74  , 70  , 66  , 62  , 59  , 56 },
	{1676, 1582, 1492, 1408, 1330, 1256, 1184, 1118, 1056, 996 , 940 , 888,
	 838 , 791 , 746 , 704 , 665 , 628 , 592 , 559 , 528 , 498 , 470 , 444,
	 419 , 395 , 373 , 352 , 332 , 314 , 296 , 280 , 264 , 249 , 235 , 222,
	 209 , 198 , 187 , 176 , 166 , 157 , 148 , 140 , 132 , 125 , 118 , 111,
	 104 , 99  , 93  , 88  , 83  , 78  , 74  , 70  , 66  , 62  , 59  , 55 },
	{1664, 1570, 1482, 1398, 1320, 1246, 1176, 1110, 1048, 990 , 934 , 882,
	 832 , 785 , 741 , 699 , 660 , 623 , 588 , 555 , 524 , 495 , 467 , 441,
	 416 , 392 , 370 , 350 , 330 , 312 , 294 , 278 , 262 , 247 , 233 , 220,
	 208 , 196 , 185 , 175 , 165 , 156 , 147 , 139 , 131 , 124 , 117 , 110,
	 104 , 98  , 92  , 87  , 82  , 78  , 73  , 69  , 65  , 62  , 58  , 55 },
	{1652, 1558, 1472, 1388, 1310, 1238, 1168, 1102, 1040, 982 , 926 , 874,
	 826 , 779 , 736 , 694 , 655 , 619 , 584 , 551 , 520 , 491 , 463 , 437,
	 413 , 390 , 368 , 347 , 328 , 309 , 292 , 276 , 260 , 245 , 232 , 219,
	 206 , 195 , 184 , 174 , 164 , 155 , 146 , 138 , 130 , 123 , 116 , 109,
	 103 , 97  , 92  , 87  , 82  , 77  , 73  , 69  , 65  , 61  , 58  , 54 },
	{1640, 1548, 1460, 1378, 1302, 1228, 1160, 1094, 1032, 974 , 920 , 868,
	 820 , 774 , 730 , 689 , 651 , 614 , 580 , 547 , 516 , 487 , 460 , 434,
	 410 , 387 , 365 , 345 , 325 , 307 , 290 , 274 , 258 , 244 , 230 , 217,
	 205 , 193 , 183 , 172 , 163 , 154 , 145 , 137 , 129 , 122 , 115 , 109,
	 102 , 96  , 91  , 86  , 81  , 77  , 72  , 68  , 64  , 61  , 57  , 54 },
	{1628, 1536, 1450, 1368, 1292, 1220, 1150, 1086, 1026, 968 , 914 , 862,
	 814 , 768 , 725 , 684 , 646 , 610 , 575 , 543 , 513 , 484 , 457 , 431,
	 407 , 384 , 363 , 342 , 323 , 305 , 288 , 272 , 256 , 242 , 228 , 216,
	 204 , 192 , 181 , 171 , 161 , 152 , 144 , 136 , 128 , 121 , 114 , 108,
	 102 , 96  , 90  , 85  , 80  , 76  , 72  , 68  , 64  , 60  , 57  , 54 },
	{1814, 1712, 1616, 1524, 1440, 1356, 1280, 1208, 1140, 1076, 1016, 960,
	 907 , 856 , 808 , 762 , 720 , 678 , 640 , 604 , 570 , 538 , 508 , 480,
	 453 , 428 , 404 , 381 , 360 , 339 , 320 , 302 , 285 , 269 , 254 , 240,
	 226 , 214 , 202 , 190 , 180 , 170 , 160 , 151 , 143 , 135 , 127 , 120,
	 113 , 107 , 101 , 95  , 90  , 85  , 80  , 75  , 71  , 67  , 63  , 60 },
	{1800, 1700, 1604, 1514, 1430, 1350, 1272, 1202, 1134, 1070, 1010, 954,
	 900 , 850 , 802 , 757 , 715 , 675 , 636 , 601 , 567 , 535 , 505 , 477,
	 450 , 425 , 401 , 379 , 357 , 337 , 318 , 300 , 284 , 268 , 253 , 238,
	 225 , 212 , 200 , 189 , 179 , 169 , 159 , 150 , 142 , 134 , 126 , 119,
	 112 , 106 , 100 , 94  , 89  , 84  , 79  , 75  , 71  , 67  , 63  , 59 },
	{1788, 1688, 1592, 1504, 1418, 1340, 1264, 1194, 1126, 1064, 1004, 948,
	 894 , 844 , 796 , 752 , 709 , 670 , 632 , 597 , 563 , 532 , 502 , 474,
	 447 , 422 , 398 , 376 , 355 , 335 , 316 , 298 , 282 , 266 , 251 , 237,
	 223 , 211 , 199 , 188 , 177 , 167 , 158 , 149 , 141 , 133 , 125 , 118,
	 111 , 105 , 99  , 94  , 88  , 83  , 79  , 74  , 70  , 66  , 62  , 59 },
	{1774, 1676, 1582, 1492, 1408, 1330, 1256, 1184, 1118, 1056, 996 , 940,
	 887 , 838 , 791 , 746 , 704 , 665 , 628 , 592 , 559 , 528 , 498 , 470,
	 444 , 419 , 395 , 373 , 352 , 332 , 314 , 296 , 280 , 264 , 249 , 235,
	 222 , 209 , 198 , 187 , 176 , 166 , 157 , 148 , 140 , 132 , 125 , 118,
	 111 , 104 , 99  , 93  , 88  , 83  , 78  , 74  , 70  , 66  , 62  , 59 },
	{1762, 1664, 1570, 1482, 1398, 1320, 1246, 1176, 1110, 1048, 988 , 934,
	 881 , 832 , 785 , 741 , 699 , 660 , 623 , 588 , 555 , 524 , 494 , 467,
	 441 , 416 , 392 , 370 , 350 , 330 , 312 , 294 , 278 , 262 , 247 , 233,
	 220 , 208 , 196 , 185 , 175 , 165 , 156 , 147 , 139 , 131 , 123 , 117,
	 110 , 104 , 98  , 92  , 87  , 82  , 78  , 73  , 69  , 65  , 61  , 58 },
	{1750, 1652, 1558, 1472, 1388, 1310, 1238, 1168, 1102, 1040, 982 , 926,
	 875 , 826 , 779 , 736 , 694 , 655 , 619 , 584 , 551 , 520 , 491 , 463,
	 437 , 413 , 390 , 368 , 347 , 328 , 309 , 292 , 276 , 260 , 245 , 232,
	 219 , 206 , 195 , 184 , 174 , 164 , 155 , 146 , 138 , 130 , 123 , 116,
	 109 , 103 , 97  , 92  , 87  , 82  , 77  , 73  , 69  , 65  , 61  , 58 },
	{1736, 1640, 1548, 1460, 1378, 1302, 1228, 1160, 1094, 1032, 974 , 920,
	 868 , 820 , 774 , 730 , 689 , 651 , 614 , 580 , 547 , 516 , 487 , 460,
	 434 , 410 , 387 , 365 , 345 , 325 , 307 , 290 , 274 , 258 , 244 , 230,
	 217 , 205 , 193 , 183 , 172 , 163 , 154 , 145 , 137 , 129 , 122 , 115,
	 108 , 102 , 96  , 91  , 86  , 81  , 77  , 72  , 68  , 64  , 61  , 57 },
	{1724, 1628, 1536, 1450, 1368, 1292, 1220, 1150, 1086, 1026, 968 , 914,
	 862 , 814 , 768 , 725 , 684 , 646 , 610 , 575 , 543 , 513 , 484 , 457,
	 431 , 407 , 384 , 363 , 342 , 323 , 305 , 288 , 272 , 256 , 242 , 228,
	 216 , 203 , 192 , 181 , 171 , 161 , 152 , 144 , 136 , 128 , 121 , 114,
	 108 , 101 , 96  , 90  , 85  , 80  , 76  , 72  , 68  , 64  , 60  , 57 }}

local ModPeriodToNoteNumber = {};
for i,v in ipairs(ModPeriodTable[1]) do
	ModPeriodToNoteNumber[v] = i
end
local function getNoteNumber(note)
	local t = ModPeriodToNoteNumber[note]
	if not t then
		local d
		for _,tbl in ipairs(ModPeriodTable) do
			for i,v in ipairs(tbl) do
				local z=math.abs(v-note)
				if d==nil or z<d then t,d=i,z end
			end
		end
		ModPeriodToNoteNumber[note]=t
	end
	return t
end

local WaveformsTable = {{},{},{},{}}
do
	local t={0,255,0,-256}
	for i=0,63 do
		WaveformsTable[1][i+1] = math.floor(math.sin(math.pi*i/32)*255)
		WaveformsTable[2][i+1] = math.floor(255-i*512/64)
		WaveformsTable[3][i+1] = t[math.floor(1+i/16)]
		WaveformsTable[4][i+1] = math.floor(math.random()*512-256)
	end
end
local VibratoTable, TremoloTable

local function clamp(x,min,max)
	if x<min then x=min end
	if x>max then x=max end
	return x
end

-- read mod file
local file = io.open(arg[1]:gsub('^/cygdrive/(%w)/','%1:/'), 'rb')
local mod = ModFile(file:read('*all'))
file:close()
if mod.channelCount>4 then error('Unsupported channel count: ' .. mod.channelCount) end

-- timing calculations 
local rate = 1000000/193
local ticksPerSecond = 7093789.2 -- PAL frequency
local ticksPerFrame = 0 -- calculated by setBpm
local ticksPerOutputSample = math.floor(.5 + ticksPerSecond / rate)
local ticksSinceStartOfFrame = 0
	
local function setBpm(bpm)
	-- x beats per minute => x*4 rows per minute 
	ticksPerFrame = math.floor(.5 + ticksPerSecond * 2.5/bpm)
end

-- initial player state 
local framesPerRow,currentFrame,currentPattern,currentPosition,currentRow,nextPosition,nextRow
local channels = {}

local function resetPlayerData()
	ticksSinceStartOfFrame = 0
	setBpm(125)
	framesPerRow = 6
	currentFrame = 0
	currentPattern = 1
	currentPosition = 1
	currentRow = 1
	nextPosition = nil
	nextRow = nil
	
	VibratoTable = WaveformsTable[1]
	TremoloTable = WaveformsTable[1]

	for chan=1,mod.channelCount do
		channels[chan] = {
			no = chan,
			playing = false,
			sample = mod.samples[1],
			finetune = 0,
			
			volume = 0,
			volumeDelta = 0,
			volumeDeltaCount = 10000, -- some big value
			
			periodDelta = 0,
			arpeggioActive = false,
			ticksPerSample = 1000000,
			
			vibrato = {
				active  = false,
				latch = 0,
				index = 0,
				speed = 0,
				depth = 0
			},
			
			tremolo = {
				active  = false,
				latch = 0,
				index = 0,
				speed = 0,
				depth = 0
			},
			
			loop = {
				start = 0,
				count = 0,
			}
		}
	end
end

local function loadRow(rowNumber) 
	thomson.ptn.add('* Row ' .. (rowNumber-1).. ' - '..(currentPattern.no-1))
	currentRow = rowNumber;
	currentFrame = 0
	for chan=1,mod.channelCount do
		local note = currentPattern[rowNumber][chan]
		local channel = channels[chan]
		-- print('* c:'..chan, 's:'..note.sample, 'f:'..note.period, 'e:'..note.effect, 'p:'..note.effectParameter)
		-- NdSam: variations spans only for a row, so reset them at beggining of a row
		local prevTicksPerSample = channel.ticksPerSample
		channel.periodDelta = 0
		channel.volumeDelta = 0
		channel.targetTicksPerSample = nil
		if note.period~=0 or note.sample~=0 then
			channel.playing = true
			channel.samplePosition = 0
			channel.ticksSinceStartOfSample = 0 -- that's 'sample' as in 'individual volume reading' 
			if note.sample ~= 0 and note.sample<=mod.sampleCount then
				channel.sample = mod.samples[note.sample]
				channel.volume = channel.sample.volume
				channel.finetune = channel.sample.finetune
				channel.tremolo.latch = channel.volume
				thomson.smp(chan).vol(chan)
			end
			if note.period ~= 0 then
				channel.noteNumber = getNoteNumber(note.period)
				channel.ticksPerSample = ModPeriodTable[1+channel.finetune][channel.noteNumber] * 2;
				channel.vibrato.latch = channel.ticksPerSample
				thomson.frq(chan)
			end
		end
		if note.effect ~= 0 or note.effectParameter ~= 0 then
			-- https://wiki.openmpt.org/Manual:_Effect_Reference
			local effectNotSupported = true

			local x = math.floor(note.effectParameter/16)
			local y = note.effectParameter%16
			
			channel.volumeDelta = 0 -- new effects cancel volumeDelta 
			channel.periodDelta = 0 -- new effects cancel periodDelta
			channel.arpeggioActive = false
			channel.vibrato.active = false
			channel.tremolo.active = false
			
			if note.effect==0x00 then -- arpeggio: 0xy
				effectNotSupported = false
				channel.arpeggioActive = true;
				channel.arpeggioNotes = {
					channel.noteNumber,
					channel.noteNumber + x,
					channel.noteNumber + y
				}
				channel.arpeggioCounter = 0;
			end
			if note.effect==0x01 then -- pitch slide up - 1xx
				effectNotSupported = false
				channel.periodDelta = -note.effectParameter
				if channel.periodDelta==0 then
					channel.periodDelta = channel.periodDeltaMem or 0
				end
				channel.periodDeltaMem = channel.periodDelta
			end
			if note.effect==0x02 then -- pitch slide down - 2xx 
				effectNotSupported = false
				channel.periodDelta = note.effectParameter
				if channel.periodDelta==0 then
					channel.periodDelta = channel.periodDeltaMem or 0
				end
				channel.periodDeltaMem = channel.periodDelta
			end
			if note.effect==0x03 or note.effect==0x05 then -- tonePortamento
				effectNotSupported = false
				if note.effectParameter==0 or note.effect==0x05 then
					note.effectParameter = channel.effectMem3 or 0
				end	
				channel.effectMem3 = note.effectParameter
				channel.targetTicksPerSample = channel.ticksPerSample
				channel.ticksPerSample = prevTicksPerSample
				channel.periodDelta = note.effectParameter
			end
			if note.effect==0x04 or note.effect==0x06 then -- Vibrato - 4xy
				effectNotSupported = false
				channel.vibrato.active = true
				channel.vibrato.speed = note.effect==0x04 and x>0 and x or channel.vibrato.speed
				channel.vibrato.depth = note.effect==0x04 and y>0 and y or channel.vibrato.depth
			end
			if note.effect==0x07 then -- Tremolo - 7xy
				effectNotSupported = false
				channel.tremolo.active = true
				channel.tremolo.speed = x>0 and x or channel.tremolo.speed
				channel.tremolo.depth = y>0 and y or channel.tremolo.depth
			end
			if note.effect==0x09 then -- sample offset
				effectNotSupported = false
				if note.sample~=0 then
					thomson.smpOffset(chan, note.sample, note.effectParameter)
				end
			end
			if note.effect==0x0a or note.effect==0x05 or note.effect==0x06 then -- volume slide - Axy
				effectNotSupported = false
				if x>0 then  -- volume increase by x 
					channel.volumeDelta = x
				else -- volume decrease by y 
					channel.volumeDelta = -y
				end
				channel.volumeDeltaCount = 10000
			end
			if note.effect==0x0b then -- position jump
				effectNotSupported = false
				nextPosition = 1 + note.effectParameter
				currentRow = 64
			end
			if note.effect==0x0c then -- volume 
				effectNotSupported = false
				channel.volume = math.min(64, note.effectParameter)
				thomson.vol(chan)
			end
			if note.effect==0x0d then -- pattern break
				effectNotSupported = false
				nextRow = 1 + note.effectParameter
				currentRow = 64
			end
			if note.effect==0x0E then -- misc
				if 0x00<=note.effectParameter and note.effectParameter<=0x01 then 
					-- filter on/off
					effectNotSupported = false
					thomson.ptn.printf('* ignoring 0xE(%d) for channel %d',note.effectParameter,chan)
				end
				if x==0x4 then
					effectNotSupported = false
					channel.vibrato.index = y<4 and 0 or channel.vibrato.index
					VibratoTable = WaveformsTable[1+(y%4)]
				end
				if x==0x5 then
					effectNotSupported = false
					if note.period>0 then
						channel.finetune = y
						channel.ticksPerSample = ModPeriodTable[1+channel.finetune][channel.noteNumber] * 2;
						channel.vibrato.latch = channel.ticksPerSample
						thomson.frq(chan)
					end
				end
				if x==0x6 then
					effectNotSupported = false
					if y==0x0 then
						channel.loop.start = currentRow
					else
						if channel.loop.count==0 then
							channel.loop.count = y
							nextRow = channel.loop.start
						else
							channel.loop.count=channel.loop.count-1
						end
					end
				end
				if x==0x7 then
					effectNotSupported = false
					channel.tremolo.index = y<4 and 0 or channel.tremolo.index
					TremoloTable = WaveformsTable[1+(y%4)]
				end				
				if x==0xA then
					effectNotSupported = false
					channel.volumeDelta = y
					channel.volumeDeltaCount = 1
				end
				if x==0xB then
					-- a bit cheating, it should occur after the first tick
					effectNotSupported = false
					channel.volumeDelta = -y
					channel.volumeDeltaCount = 1
				end
			end
			if note.effect==0x0f then -- tempo change 
				effectNotSupported = false
				if note.effectParameter == 0 then
				elseif note.effectParameter <= 32 then
					framesPerRow = note.effectParameter
				else
					setBpm(note.effectParameter)
				end
			end
			if effectNotSupported then
				local msg = string.format('* Unsupported effect 0x%02X(%02X) for Channel %d, Row %d, Pattern %d', note.effect, note.effectParameter, chan, rowNumber-1, (mod.positions[currentPosition] or mod.positions[1])-1)
				thomson.ptn.add(msg)
				-- error(msg)
			end
		end
	end
end

local function loadPattern(patternNumber)
	thomson.ptn.newPattern(patternNumber)
	
	currentPattern = mod.patterns[patternNumber]
	loadRow(nextRow or 1)
end
	
local function loadPosition(positionNumber)
	currentPosition = positionNumber;
	loadPattern(mod.positions[currentPosition] or mod.positions[1])
end

-- loadPosition(1)

local function getNextPosition()
	if currentPosition == mod.positionCount then
		loadPosition(mod.positionLoopPoint)
		thomson.playing = false
	else
		loadPosition(nextPosition or currentPosition + 1)
	end
	nextPosition = nil
end

local function getNextRow()
	if currentRow == 64 then
		getNextPosition()
	else
		loadRow(nextRow or currentRow + 1)
	end
	nextRow = nil
end

local function doFrame()
	currentFrame = currentFrame+1
	-- thomson.ptn.add('* frame'..currentFrame)
	-- apply volume/pitch slide before fetching row, because the first frame of a row does NOT have the slide applied 
	for chan=1,mod.channelCount do
		local channel = channels[chan]
		if channel.volumeDeltaCount>0 then
			channel.volumeDeltaCount = channel.volumeDeltaCount-1
			channel.volume = clamp(channel.volume + channel.volumeDelta,0,64)
		end
		if channel.targetTicksPerSample then
			channel.periodDelta = math.abs(channel.periodDelta)*(channel.ticksPerSample<channel.targetTicksPerSample and 1 or -1)
		end
		channel.ticksPerSample = clamp(channel.ticksPerSample + channel.periodDelta * 2,96,4096)
		if channel.arpeggioActive then
			local noteNumber = channel.arpeggioNotes[1+(channel.arpeggioCounter % 3)]
			channel.ticksPerSample = ModPeriodTable[1+channel.finetune][math.min(noteNumber,#ModPeriodTable[1])] * 2;
			channel.arpeggioCounter = channel.arpeggioCounter+1
		end
		if channel.vibrato.active then
			local e = channel.vibrato
			channel.ticksPerSample = clamp(e.latch + math.floor(VibratoTable[1 + e.index]*e.depth/127),96,4096)
			e.index  = (e.index + e.speed)%64
		end
		if channel.tremolo.active then
			local e = channel.tremolo
			channel.volume = clamp(e.latch + math.floor(TremoloTable[1 + e.index]*e.depth/127),0,64)
			e.index  = (e.index + e.speed)%64
		end
		thomson.vol(chan).frq(chan)
	end
	
	if currentFrame == framesPerRow then
		getNextRow();
	end
end

----------------------------------------------------------------------------------------------

function patternSection()
	-- optimizing pattern sections inspire by LZW
	-- https://fr.wikipedia.org/wiki/Lempel-Ziv-Welch
	
	local ps = {
		record=true,
		compress=thomson.compress,
		dict={['']={
			count=0,
			add=function(self) end,
			call=function(self) return {} end,
			inline=function(self) return {} end
		}},
		libNo=0,
		W='',
		wait={},
		defined={}
	}
	
	function ps.printTable(tbl,format)
		format=format or "$%x"
		local t,s='',' fcb '
		for _,v in ipairs(tbl) do
			if type(v)~="number" then
				print(t..v)
				t = ''; s=' fcb '
			else
				local a=string.format(format, v)
				local b=t..s..a; s=','
				if b:len()>=40 then
					print(t)
					b = ' fcb '..a
				end
				t=b
			end
		end
		if t:len()>0 then print(t) end
	end
	
	function ps.libName(no)
		return string.format('LIB%04d',no)
	end
	
	function ps.add(arg)
		if ps.record then 
			if ps.compress then
				ps.addCompress(arg)
			else
				ps.addNoCompress(arg)
			end
		end
	end
	
	function ps.addNoCompress(arg)
		arg.call=function(self) return self end
		table.insert(ps, arg)
	end
	
	function ps.newLib(lib,code)
		lib = lib or ps.dict['']
		if code == nil then error('nil') end
		lib.count=lib.count+1
		ps.libNo = ps.libNo+1
		return {
			no = ps.libNo,
			lib = lib,
			code = code,
			count = 0,
			called = false,
			add = function(self)
				table.insert(ps, self)
				self.count = self.count + 1
			end,
			inline = function(self)
				local t = {}
				for _,v in ipairs(self.lib:call()) do table.insert(t,v) end
				for _,v in ipairs(self.code) do table.insert(t, v) end
				return t
			end,
			call = function(self)
				if self.called then
					return self.called
				else
					function size(code)
						local t=0
						for _,v in ipairs(code) do
							if type(v)=="string" then
								if v:match(' fdb ') then t=t+2 end
							else
								t=t+1
							end
						end
						return t
					end
				
					local inline = self:inline()
					local inline_size = size(inline)
					local gain = self.count*inline_size - (2*self.count + inline_size + 1)
					if gain<=0 then
						return inline
					else
						self.called = {' fdb ' ..ps.libName(self.no)..'-song'}
						return self.called
					end
				end
			end
		}
	end
		
	function ps.addCompress(arg)
		if type(arg)=="table" then
			local key = ''
			for _,v in ipairs(arg) do
				if type(v)=="number" then 
					key=key..string.char(v) 
					-- key=key..v..'|'
				end
			end
			if ps.dict[ps.W..key] then
				ps.W = ps.W .. key
			else
				local lib =	ps.dict[ps.W]
				if not lib then error(ps.W) end
				
				ps.dict[ps.W .. key] = ps.newLib(lib,arg)
			
				lib:add()
				ps.W = key
				if not ps.dict[ps.W] then
					ps.dict[ps.W] = ps.newLib(nil,arg)
				end
			end
		else
			ps.dict[ps.W]:add()
			ps.W=''
			table.insert(ps, arg)		
		end
	end
	
	function ps.printWait()
		local created={}
		while true do
			local max=0
			for i in pairs(ps.wait) do
				max=math.max(i,max)
			end
			if max==0 then break end
			if not created[max] then
				while max>3*15 do
					if not created[max] then
						created[max]=true
						ps.wait[max]=nil
						print('WAIT'..max)
					else
						print(' fdb WAIT'..max..'-song')
						break
					end
					if max%2>0 then
						local t=max%16
						print(string.format(' fcb $8%x',t))
						max=max-t
					end
					max=math.floor(max/2)
					print(' fdb WAIT'..max..'-song')
				end
				if max<=3*15 then
					if created[max] then
						print(' fdb WAIT'..max..'-song')
					else
						created[max]=true
						ps.wait[max]=nil
						print('WAIT'..max)
						if max>15 then print(' fcb $8f'); max=max-15 end
						if max>15 then print(' fcb $8f'); max=max-15 end
						print(string.format(' fcb $8%x',max))
					end
				end
				print(' fcb 0')
			end
		end
	end
	
	function ps.print()
		for _,v in ipairs(ps) do
			if type(v)=="table" then
				ps.printTable(v:call())
			else
				print(v)
			end
		end
		ps.printWait()
		
		local lib = {}
		for _,v in pairs(ps.dict) do
			if v.called then table.insert(lib,v) end
		end
		table.sort(lib, function(a,b) return a.no<b.no end)
		for _,v in ipairs(lib) do
			print(ps.libName(v.no))
			ps.printTable(v.lib:call())
			ps.printTable(v.code)
			print(' fcb 0')
		end
	end

	
	function ps.patternName(no)
		return string.format("PAT%02d",no-1)
	end
	
	function ps.newPattern(no)
		local function closePreviousPattern()
			thomson.delay(0)
			ps.add(' fcb 0') 
		end
		if not ps.defined[no] then
			ps.record = true
			ps.defined[no] = true
			if #ps>0 then 
				closePreviousPattern()
			end
			ps.add(ps.patternName(no))
		else
			closePreviousPattern()
			ps.record = false
		end
	end
	
	function ps.addWait(delay)
		local l={}
		while delay>0 do
			local t=delay>15 and 15 or delay
			delay = delay - t
			ps.add({128+t,'\t; d:'..t})
		end
	end

    function ps.addWait_(delay)
        if delay<=2*15 then
            local comment = ' ; d:'..delay
            local l={}
            while delay>0 do
				local t=delay>15 and 15 or delay
				delay = delay - t
				table.insert(l,128+t)
			end
			table.insert(l,comment)
			ps.add(l)
		end
		if delay>0 then
			ps.wait[delay] = true
			ps.add(' fdb WAIT'..delay..'-song')			
		end
	end

	function ps.printf(...)
		ps.add(string.format(...))
	end
	
	function ps.clear()
		for i in ipairs(ps) do
			ps[i] = nil
		end
		for v in pairs(ps.dict) do
			if v:len()>0 then ps.dict[v]=nil end
		end
		ps.W=''
	end
	
	return ps
end

----------------------------------------------------------------------------------------------

function none(arg) end

function thomson.reset()
	thomson.chl = {}
	for i=1,4 do
		thomson.chl[i] = {prev={vol=nil,frq=nil,smp=nil},
						  curr={vol=nil,frq=nil,smp=nil}}
	end
	thomson.cur = 0
	thomson.dly = 0
	thomson.smp2 = thomson.smp2 or {}
	
	thomson.ptn = patternSection()
	
	return thomson
end
function thomson.vol(chan)
	thomson.chl[chan].curr.vol = math.ceil(channels[chan].volume*15/64)
	return thomson
end
function thomson.frq(chan)
	-- if(thomson.chl[chan].curr.frq==209) then error('xx') end
	thomson.chl[chan].curr.frq = math.floor(.5 + 256*ticksPerOutputSample/(channels[chan].ticksPerSample*channels[chan].sample.thomson.div)) 
	return thomson
end
function thomson.smp(chan)
	thomson.chl[chan].prev.smp = nil
	thomson.chl[chan].curr.smp = channels[chan].sample.no
	return thomson
end
function thomson.smpOffset(chan,no,offset)
	thomson.chl[chan].prev.smp = nil
	local t=no*256+offset
	if not thomson.smp2[t] then
		local free
		for i=1,mod.sampleCount do
			if mod.samples[i].length==0 then
				free=i
				break;
			end
		end
		if free==nil then
			mod.sampleCount = mod.sampleCount+1
			if mod.sampleCount>256 then
				error('No more free samples for offset')
			end
			free = mod.sampleCount				
		end
		thomson.smp2[t] = free
		local name = mod.samples[free] and mod.samples[free].name
		mod.samples[free] = {
		    no       = free-1,
			source   = mod.samples[no],
			offset   = offset*256,
			volume   = mod.samples[no].volume,
			finetune = mod.samples[no].finetune,
			thomson  = mod.samples[no].thomson
		}
		mod.samples[free].name = (name and name..' ' or '')..' s:'..no..'+'..offset
	end
	thomson.chl[chan].curr.smp = thomson.smp2[t]-1
	return thomson
end
function thomson.delay(dly)
	local changes = {}
	-- thomson.debug('* delay',dly,thomson.dly)
	for chl=1,4 do
		local vol = thomson.chl[chl].curr.vol
		local frq = thomson.chl[chl].curr.frq
		local smp = thomson.chl[chl].curr.smp
		local volChg = thomson.chl[chl].prev.vol ~= vol
		local frqChg = thomson.chl[chl].prev.frq ~= frq
		local smpChg = thomson.chl[chl].prev.smp ~= smp
		-- thomson.debug('* c:'..chl,volChg,frqChg,smpChg)
		if volChg or frqChg or smpChg then
			if chl~=thomson.cur or volChg then
				thomson.cur = chl
				thomson.chl[chl].prev.vol = vol
				table.insert(changes, {(3*4 + chl-1)*16+vol, '\t; c:'..chl..' v:'..vol})
			end
			if smpChg then
				thomson.chl[chl].prev.smp = smp
				if smp>=31 then	
					table.insert(changes, {0xA0, smp%256, '\t; s:'..(smp+1)})
				else
					table.insert(changes, {0xA1 +    smp, '\t; s:'..(smp+1)})
				end
			end
			if frqChg then
				thomson.chl[chl].prev.frq = frq
				local hi,lo = math.floor(frq/256),frq%256
				if hi>15 then
					table.insert(changes, {0x90,  hi,lo,'\t; f:'..frq})
				else
					table.insert(changes, {0x91 + hi,lo,'\t; f:'..frq})
				end
			end

			-- stats
			channels[chl].sample.thomson.frq = channels[chl].sample.thomson.frq or {min=frq,max=frq}
			channels[chl].sample.thomson.frq.min = math.min(frq, channels[chl].sample.thomson.frq.min)
			channels[chl].sample.thomson.frq.max = math.max(frq, channels[chl].sample.thomson.frq.max)
		end
	end

	if dly==-1 then -- reset state
		dly = 0
		thomson.cur = 0
		for chl=1,4 do
			thomson.chl[chl].prev.vol = nil
			thomson.chl[chl].prev.frq = nil
			thomson.chl[chl].prev.smp = nil
		end
	end
	
	local flush = dly==0 or #changes>0 -- or true
	if flush and thomson.dly>0 then
		thomson.ptn.addWait(thomson.dly)
		thomson.dly=0
	end
	
	for _,v in ipairs(changes) do 
		thomson.ptn.add(v) 
	end

	thomson.dly=thomson.dly+dly
	
	return thomson
end

local function play()
	resetPlayerData()

	loadPosition(1)
	
	thomson.playing = true
	while thomson.playing do
		local delay=0
		while ticksSinceStartOfFrame<ticksPerFrame do
			ticksSinceStartOfFrame = ticksSinceStartOfFrame + ticksPerOutputSample*16;
			delay=delay + 1
		end
		while ticksSinceStartOfFrame >= ticksPerFrame do
			doFrame();
			ticksSinceStartOfFrame = ticksSinceStartOfFrame - ticksPerFrame;
		end
		thomson.delay(delay)
	end
	thomson.delay(0)
end


-- print('ticksPerFrame='..ticksPerFrame)
-- print('ticksPerOutputSample='..ticksPerOutputSample)


print('* MODULE: '..mod.title)
print('* Identifier: ' .. mod.identifier)
print('* Channels  : ' .. mod.channelCount)
print('* Samples   : ' .. mod.sampleCount)
print('* Patterns  : ' .. mod.patternCount)
print('* Positions : ' .. mod.positionCount)
print('song fdb instru-song')

-- collect statistics for optimizations
thomson.reset()
for _,sample in ipairs(mod.samples) do 
	sample.thomson = {div=thomson.div} 
end	
play()

-- do the real translation
thomson.reset()
for i=1,4 do
	channels[i].sample = mod.samples[1]
	channels[i].volume = 0
	thomson.vol(i).smp(i).frq(i)
end
thomson.delay(-1)
if false then -- test: play instru indiviually
	for i=1,mod.sampleCount do
		local channel = channels[1]
		channel.sample = mod.samples[i]
		channel.volume = 64
		channel.ticksPerSample = ModPeriodTable[1][12*3]
		thomson.smp(1).vol(1).frq(1).delay(200)
		channel.volume = 0
		thomson.smp(1).vol(1).frq(1).delay(200).delay(-1)

	end
	thomson.ptn.add(' fcb 0')
end
thomson.ptn.print()
thomson.ptn.clear()
for i=1,mod.positionCount do
	print(' fdb '..thomson.ptn.patternName(mod.positions[i])..'-song')
end
print('* repeat ?')
for i=mod.positionLoopPoint,mod.positionCount do
	print(' fdb '..thomson.ptn.patternName(mod.positions[i])..'-song')
end
print(' fcb 0')

print('instru')
local norm = 1/128;
if false then -- uses RMS to get louder sound
	local v,n=0,0
	for i=1,mod.sampleCount do
		local sample = mod.samples[i]
		if not sample.source then
			for p=0,sample.length-1 do
				v = v + sample.getSample(p)^2
				n = n + 1
			end
		end
	end
	local rms = (v/n)^.5
	norm = .707/rms 
		   -- (3/4)^.5/rms
end

-- fill data for non derived samples
dat = {}
local function buildSample(sample)
	-- table.insert(dat,'* s:'..sample.no)
	sample.thomson.start = #dat
	sample.thomson.len   = 0
	sample.thomson.rept  = nil
	
	if sample.thomson.frq then --- played sample
		if false and sample.thomson.frq.max>255 then
			if true then
				local k = sample.thomson.frq.max/256
				sample.thomson.div = sample.thomson.div*k
				sample.thomson.frq.min = sample.thomson.frq.min/k
				sample.thomson.frq.max = sample.thomson.frq.max/k
			else
				while sample.thomson.frq.max>255 do
					sample.thomson.frq.max = sample.thomson.frq.max/2
					sample.thomson.frq.min = sample.thomson.frq.min/2
					sample.thomson.div = sample.thomson.div*2
				end
			end
			print('* Div for sample ',sample.no,'=',sample.thomson.div, sample.thomson.frq.max)
		end
		
		local length = sample.length
		if sample.repeatOffset>0 then 
			length = sample.repeatOffset + sample.repeatLength 
		end
		
		do
			local t=sample.thomson.div
			while sample.thomson.div>1 and length<=256*sample.thomson.div do
				sample.thomson.div = sample.thomson.div/2
			end

			while length>=4096*sample.thomson.div do
				sample.thomson.div = sample.thomson.div*2
			end
			
			if sample.thomson.div~=t then
				print('* Changing div for sample ',sample.no,' from ',t,' to ',sample.thomson.div)
			end
		end
		
		local ok,p,v,d=true,0,0,0
		while ok do
			if sample.repeatOffset>0 and not sample.thomson.rept and p==sample.repeatOffset then
				sample.thomson.rept=sample.thomson.len
				-- print('* repeat')
			end
			local b = sample.getSample(p)*norm
			d = d + 1
			if d<sample.thomson.div then
				v = v + b
			else
				d=d-sample.thomson.div
				v = (v + b*(1-d))/sample.thomson.div 
				-- print('* sample:',i,p,v,d,f)
				table.insert(dat, math.floor(.5+7.5*(1+clamp(v,-1,1))))
				sample.thomson.len = sample.thomson.len + 1
				v = b*d
				if p>=length then ok = false end
			end
			p = p + 1
			if p==length and sample.repeatOffset>0 and sample.thomson.len<256
			then 
				p=sample.repeatOffset
			end
		end
	end
	
	sample.thomson.stop=sample.thomson.start+sample.thomson.len
	sample.thomson.rept=sample.thomson.rept and sample.thomson.start+sample.thomson.rept
											 or sample.thomson.stop
end
local function buildSample2(sample)
	sample.thomson.start = #dat
	sample.thomson.stop  = #dat
	sample.thomson.rept  = #dat
	sample.thomson.len   = 0
	
	if sample.thomson.frq then --- played sample
		if false and sample.thomson.frq.max>255 then
			if true then
				local k = sample.thomson.frq.max/255
				sample.thomson.div = sample.thomson.div*k
				sample.thomson.frq.min = sample.thomson.frq.min/k
				sample.thomson.frq.max = sample.thomson.frq.max/k
			else
				while sample.thomson.frq.max>255 do
					sample.thomson.frq.max = sample.thomson.frq.max/2
					sample.thomson.frq.min = sample.thomson.frq.min/2
					sample.thomson.div = sample.thomson.div*2
				end
			end
			print('* Div for sample ',sample.no,'=',sample.thomson.div, sample.thomson.frq.max)
		end
	
		local len = math.floor(.5+sample.length/sample.thomson.div)
		local rpt = math.floor(.5+sample.repeatOffset/sample.thomson.div)
		local rle = math.floor(.5+sample.repeatLength/sample.thomson.div)
		
		local function getSample(pos)
			local t,p=0,.5+(pos-1)*sample.thomson.div
			local div = math.ceil(sample.thomson.div)
			-- div = 1 -- only one sample
			for j=1,div do
				while p>=sample.length do
					p = sample.repeatLength>0 and p-sample.repeatLength or p-1
				end
				t = t + sample.getSample(math.floor(p))*norm
				p=p+1
			end
			t=t/div
			return math.floor(.5+7.5*(1+clamp(t,-1,1)))
		end
		
		-- iQf i==1 --or i==3 -- or i==4 -- or i==5 
		-- then len,rpt,rle = 0,0,0 end
		
		local ext = 0
		if rpt==0 then 
			rpt=len
			rle=0
		else
			if rpt+rle>len then
				rle = len-rpt
				-- if rle<0 then rle = 0 end
			end
			if rpt+rle<len then
				len = rpt+rle
			end
			local t=rle
			while len<256 do
				len = len+t
				rle = rle+t
			end
		end
		sample.thomson.stop=sample.thomson.stop+len+ext
		sample.thomson.rept=sample.thomson.rept+rpt
		sample.thomson.len=len
		for i=1,len do
			table.insert(dat, getSample(i))
		end
		-- add extra to cope with overflow
		for i=1,ext do table.insert(dat,dat[#dat]) end
	end
end
for i=1,mod.sampleCount do
	local sample = mod.samples[i]
	if not sample.source then
		buildSample(sample)
		
		if sample.thomson.rept>=sample.thomson.stop-1 then
			-- optim: remove tailing part if constant
			local last = #dat
			while sample.thomson.stop>sample.thomson.start and dat[last-1]==dat[last] do
				sample.thomson.stop = sample.thomson.stop - 1
				sample.thomson.rept = sample.thomson.rept - 1
				dat[last] = nil
				last = last-1
			end
		end
	end
end
-- fill thomson info for derived
for i=1,mod.sampleCount do
	local sample = mod.samples[i]
	if sample.source then
		local source = sample.source
		local delta  = math.floor(sample.offset/source.thomson.div)
		sample.thomson = {start = source.thomson.start+delta,
		                  stop  = source.thomson.stop,
						  rept  = source.thomson.rept,
						  div   = source.thomson.div}
		if sample.thomson.start > sample.thomson.stop then
			sample.thomson.start = sample.thomson.rept
		end
		sample.thomson.len = sample.thomson.stop-sample.thomson.start
	end
end
-- output sample info
for i=1,mod.sampleCount do
	local sample = mod.samples[i]
	print(string.format(' fdb %-2d ; len=%-4d     : %s',i,sample.thomson.len,sample.name))
	print(string.format(' fdb sample-song+%d',sample.thomson.start))
	print(string.format(' fdb sample-song+%d',sample.thomson.stop))
	print(string.format(' fdb sample-song+%d',sample.thomson.rept))
end

print(' IFGE song+256-*')
print('   ORG song+256')
print(' ENDC')

-- reset freq stats
for _,sample in ipairs(mod.samples) do 
	sample.thomson.frq = nil
end	
play()
thomson.ptn.print()

-- non played patterns
for i=1,mod.positionCount do
	if not thomson.ptn.defined[mod.positions[i]] then
		print(thomson.ptn.patternName(mod.positions[i]))
		print(' fcb 0')
	end
end

-- output sample data
print('sample')
thomson.ptn.printTable(dat)

print('* END OF MODULE: '..mod.title)
